@@ -1,7 +1,7 @@
 /*************************************************************************
  *
  * ENVMODULES.C, Modules Tcl extension library
- * Copyright (C) 2018-2019 Xavier Delaruelle
+ * Copyright (C) 2018-2020 Xavier Delaruelle
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,10 +24,23 @@
 #include <limits.h>
 #include <sys/types.h>
 #include <dirent.h>
+#include <pwd.h>
+#include <grp.h>
 #include <string.h>
+#include <stdlib.h>
 #include "config.h"
 #include "envmodules.h"
 
+
+/*	Utility function to compare 2 integers for qsort function. */
+int
+__Envmodules_IntCmp(
+   const void* i,
+   const void* j)
+{
+   return (i > j) - (i < j);
+}
+
 /*----------------------------------------------------------------------
  *
  * Envmodules_GetFilesInDirectoryObjCmd --
@@ -51,7 +64,6 @@ Envmodules_GetFilesInDirectoryObjCmd(
    int objc,               /* Number of arguments. */
    Tcl_Obj *const objv[])  /* Argument objects. */
 {
-   int fetch_hidden;
    int fetch_dotversion;
    const char *dir;
    int dirlen;
@@ -60,22 +72,18 @@ Envmodules_GetFilesInDirectoryObjCmd(
    struct dirent *direntry;
    int have_modulerc = 0;
    int have_version = 0;
+   int is_hidden;
    char path[PATH_MAX];
 
    /* Parse arguments. */
-   if (objc == 4) {
-      /* fetch_hidden */
-      if (Tcl_GetBooleanFromObj(interp, objv[2], &fetch_hidden) != TCL_OK) {
-         Tcl_SetErrorCode(interp, "TCL", "VALUE", "BOOLEAN", NULL);
-         return TCL_ERROR;
-      }
+   if (objc == 3) {
       /* fetch_dotversion */
-      if (Tcl_GetBooleanFromObj(interp, objv[3], &fetch_dotversion)!=TCL_OK) {
+      if (Tcl_GetBooleanFromObj(interp, objv[2], &fetch_dotversion)!=TCL_OK) {
          Tcl_SetErrorCode(interp, "TCL", "VALUE", "BOOLEAN", NULL);
          return TCL_ERROR;
       }
    } else {
-      Tcl_WrongNumArgs(interp, 1, objv, "dir fetch_hidden fetch_dotversion");
+      Tcl_WrongNumArgs(interp, 1, objv, "dir fetch_dotversion");
       return TCL_ERROR;
    }
 
@@ -111,15 +119,10 @@ Envmodules_GetFilesInDirectoryObjCmd(
          if (fetch_dotversion && !access(path, R_OK)) {
             have_version = 1;
          }
-      } else if (direntry->d_name[0] == '.') {
-         /* add hidden file if enabled */
-         if (fetch_hidden) {
-            Tcl_ListObjAppendElement(interp, ltmp, Tcl_NewStringObj(path, -1));
-            Tcl_ListObjAppendElement(interp, ltmp, Tcl_NewIntObj(1));
-         }
       } else {
          Tcl_ListObjAppendElement(interp, ltmp, Tcl_NewStringObj(path, -1));
-         Tcl_ListObjAppendElement(interp, ltmp, Tcl_NewIntObj(0));
+         is_hidden = (direntry->d_name[0] == '.') ? 1 : 0;
+         Tcl_ListObjAppendElement(interp, ltmp, Tcl_NewIntObj(is_hidden));
       }
    }
    /* Do not treat error happening during read to send list of valid files. */
@@ -257,6 +260,179 @@ Envmodules_ReadFileObjCmd(
    return TCL_OK;
 }
 
+/*----------------------------------------------------------------------
+ *
+ * Envmodules_InitStateUsernameObjCmd --
+ *
+ *	 This function is invoked to return the username of user running
+ *	 current process.
+ *
+ * Results:
+ *	 A standard Tcl result.
+ *
+ * Side effects:
+ *	 None.
+ *
+ *---------------------------------------------------------------------*/
+
+int
+Envmodules_InitStateUsernameObjCmd(
+   ClientData dummy,       /* Not used. */
+   Tcl_Interp *interp,     /* Current interpreter. */
+   int objc,               /* Number of arguments. */
+   Tcl_Obj *const objv[])  /* Argument objects. */
+{
+   uid_t uid;
+   struct passwd *pwd;
+   char uidstr[16];
+   Tcl_Obj *res;
+
+   /* Get current user id */
+   uid = getuid();
+
+   /* Fetch corresponding passwd entry */
+   if ((pwd = getpwuid(uid)) == NULL) {
+      Tcl_SetErrno(errno);
+      sprintf (uidstr, "%d", uid);
+#if TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION < 5
+      Tcl_AppendResult(interp, "couldn't find name for user id \"", uidstr,
+         "\": ", Tcl_PosixError(interp), (char *) NULL);
+#else
+      Tcl_SetObjResult(interp,
+         Tcl_ObjPrintf("couldn't find name for user id \"%s\": %s", uidstr,
+         Tcl_PosixError(interp)));
+#endif
+      return TCL_ERROR;
+   }
+
+   /* Set username as result */
+   res = Tcl_NewObj();
+   Tcl_IncrRefCount(res);
+   Tcl_AppendToObj(res, pwd->pw_name, strlen(pwd->pw_name));
+
+   Tcl_SetObjResult(interp, res);
+   Tcl_DecrRefCount(res);
+   return TCL_OK;
+}
+
+/*----------------------------------------------------------------------
+ *
+ * Envmodules_InitStateUsergroupsObjCmd --
+ *
+ *	 This function is invoked to return all the groups the user running
+ *	 current process is member of.
+ *
+ * Results:
+ *	 A standard Tcl result.
+ *
+ * Side effects:
+ *	 None.
+ *
+ *---------------------------------------------------------------------*/
+
+int
+Envmodules_InitStateUsergroupsObjCmd(
+   ClientData dummy,       /* Not used. */
+   Tcl_Interp *interp,     /* Current interpreter. */
+   int objc,               /* Number of arguments. */
+   Tcl_Obj *const objv[])  /* Argument objects. */
+{
+   int maxgroups;
+   GETGROUPS_T *groups;
+   int ngroups = 0;
+   int egid_in_groups = 0;
+   GETGROUPS_T egid;
+   int i, j;
+   struct group *grp;
+   char gidstr[16];
+   Tcl_Obj *lres;
+
+   /* Get actually configured number of groups */
+#if defined(HAVE_SYSCONF) && defined(_SC_NGROUPS_MAX)
+   maxgroups = sysconf(_SC_NGROUPS_MAX);
+#else
+#  if defined(NGROUPS_MAX)
+   maxgroups = NGROUPS_MAX;
+#  else
+   maxgroups = DEFAULT_MAXGROUPS;
+#  endif
+#endif
+
+   /* Fetch supplementary group list unless getgroups not supported */
+   groups = (GETGROUPS_T *) ckalloc(maxgroups * sizeof(GETGROUPS_T));
+
+#if defined (HAVE_GETGROUPS)
+   if ((ngroups = getgroups(maxgroups, groups)) == -1) {
+      Tcl_SetErrno(errno);
+#if TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION < 5
+      Tcl_AppendResult(interp, "couldn't get supplementary groups: ",
+         Tcl_PosixError(interp), (char *) NULL);
+#else
+      Tcl_SetObjResult(interp,
+         Tcl_ObjPrintf("couldn't get supplementary groups: %s",
+         Tcl_PosixError(interp)));
+#endif
+      ckfree((char *) groups);
+      return TCL_ERROR;
+   }
+#endif
+
+   /* Sort then remove duplicates from getgroups result */
+   if (ngroups > 1) {
+      qsort(groups, ngroups, sizeof(GETGROUPS_T), __Envmodules_IntCmp);
+      j = 0;
+      for (i = 1; i < ngroups; i++) {
+         if (groups[i] != groups[j]) {
+            j++;
+            groups[j] = groups[i];
+         }
+      }
+      ngroups = j + 1;
+   }
+
+   /* Add primary group if not part of getgroups result (or if getgroups
+    * function is not available) */
+   egid = getegid();
+   for (i = 0; i < ngroups; i++) {
+      if (egid == groups[i]) {
+         egid_in_groups = 1;
+         break;
+      }
+   }
+   if (egid_in_groups == 0) {
+      groups[ngroups] = egid;
+      ngroups++;
+   }
+
+   /* Add group name of primary gid and each supplementatry gid to result
+    * list */
+   lres = Tcl_NewObj();
+   Tcl_IncrRefCount(lres);
+   for (i = 0; i < ngroups; i++) {
+      if ((grp = getgrgid(groups[i])) == NULL) {
+         Tcl_SetErrno(errno);
+         sprintf(gidstr, "%d", groups[i]);
+#if TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION < 5
+         Tcl_AppendResult(interp, "couldn't find name for group id \"",
+            gidstr, "\": ", Tcl_PosixError(interp), (char *) NULL);
+#else
+         Tcl_SetObjResult(interp,
+            Tcl_ObjPrintf("couldn't find name for group id \"%s\": %s",
+            gidstr, Tcl_PosixError(interp)));
+#endif
+         ckfree((char *) groups);
+         return TCL_ERROR;
+      }
+      Tcl_ListObjAppendElement(interp, lres, Tcl_NewStringObj(grp->gr_name,
+         -1));
+   }
+
+   Tcl_SetObjResult(interp, lres);
+   Tcl_DecrRefCount(lres);
+   ckfree((char *) groups);
+   return TCL_OK;
+}
+
 /*----------------------------------------------------------------------
  *
  * Envmodules_Init --
@@ -286,6 +462,12 @@ Envmodules_Init(
    Tcl_CreateObjCommand(interp, "getFilesInDirectory",
       Envmodules_GetFilesInDirectoryObjCmd, (ClientData) NULL,
       (Tcl_CmdDeleteProc*) NULL);
+   Tcl_CreateObjCommand(interp, "initStateUsername",
+      Envmodules_InitStateUsernameObjCmd, (ClientData) NULL,
+      (Tcl_CmdDeleteProc*) NULL);
+   Tcl_CreateObjCommand(interp, "initStateUsergroups",
+      Envmodules_InitStateUsergroupsObjCmd, (ClientData) NULL,
+      (Tcl_CmdDeleteProc*) NULL);
 
    /* Provide the Envmodules package */
    return Tcl_PkgProvide(interp, PACKAGE_NAME, PACKAGE_VERSION);
